#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stralloc.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>
#include <afblib/inbuf.h>
#include <afblib/inbuf_sareadline.h>
#include <afblib/outbuf.h>
#include <afblib/outbuf_printf.h>
#include <afblib/strlist.h>
#include <afblib/tokenizer.h>
#include "pipecmd.h"

static volatile sig_atomic_t interrupted = 0;

void interrupt_handler(int sig) {
   interrupted = 1;
}

inbuf in = {0};
outbuf out = {1};
outbuf err = {2};

/* print error much like perror but by using an outbuf */
static void print_error(const char* s) {
   outbuf_printf(&err, "%s: %s\n", s, strerror(errno));
   outbuf_flush(&err);
}

static void print_child_status(pid_t pid, int wstat) {
   outbuf_printf(&out, "[%d] ", (int) pid);
   if (WIFEXITED(wstat)) {
      outbuf_printf(&out, "exit %d", WEXITSTATUS(wstat));
   } else if (WIFSIGNALED(wstat)) {
      outbuf_printf(&out, "terminated with signal %d", WTERMSIG(wstat));
      if (WCOREDUMP(wstat)) outbuf_printf(&out, " (core dump)");
   } else if (WIFSTOPPED(wstat)) {
      outbuf_printf(&out, "stopped with signal %d", WSTOPSIG(wstat));
   } else if (WIFCONTINUED(wstat)) {
      outbuf_printf(&out, "continued");
   } else {
      outbuf_printf(&out, "???");
   }
   outbuf_printf(&out, "\n");
}

static bool getline(stralloc* line) {
   outbuf_printf(&out, "pipesh> ");
   outbuf_flush(&out);
   return inbuf_sareadline(&in, line);
}

int main() {
   struct sigaction action = {
      .sa_handler = interrupt_handler,
      .sa_flags = SA_RESTART,
   };
   if (sigaction(SIGINT, &action, 0) != 0 ||
         sigaction(SIGQUIT, &action, 0) != 0) {
      print_error("sigaction");
   }

   stralloc line = {0}; strlist tokens = {0};
   while (getline(&line)) {
      stralloc_0(&line); /* required by tokenizer() */
      if (!tokenizer(&line, &tokens)) break;
      if (tokens.len == 0) continue;
      pipecmd* command = pipe_constructor(&tokens);
      if (command == 0) continue;

      pipecmd* lastp = 0; int write_fd = 1;
      for (pipecmd* p = command; p->input; p = p->input) {
	 if (!pconnect2(p->cmdname, p->argv->list,
	       PIPE_WRITE, write_fd, &p->pipe)) {
	    print_error(p->cmdname); exit(255);
	 }
	 if (lastp) {
	    pcut(&lastp->pipe);
	 }
	 lastp = p;
	 pshare(&p->pipe); write_fd = p->pipe.fd;
      }
      pid_t child = fork();
      if (child < 0) {
	 print_error("fork"); exit(255);
      }
      if (child == 0) {
	 pipecmd* p;
	 if (lastp) {
	    dup2(write_fd, 1); close(write_fd);
	    p = lastp->input;
	 } else {
	    p = command;
	 }
	 execvp(p->cmdname, p->argv->list);
	 print_error(p->cmdname); exit(255);
      }
      if (lastp) {
	 pcut(&lastp->pipe);
      }
      for (pipecmd* p = command; p->input; p = p->input) {
	 if (pwait(&p->pipe)) {
	    int wstat = p->pipe.wstat;
	    if (!WIFEXITED(wstat) || WEXITSTATUS(wstat)) {
	       print_child_status(p->pipe.pid, wstat);
	    }
	 }
      }
      pipe_free(command);

      /* wait for termination of child */
      int wstat;
      pid_t pid = wait(&wstat);
      if (pid == child) {
	 if (!WIFEXITED(wstat) || WEXITSTATUS(wstat)) {
            print_child_status(pid, wstat);
         }
      } else {
         print_error("wait");
      }
   }
}
